/**
 * This work was created by participants in the DataONE project, and is
 * jointly copyrighted by participating institutions in DataONE. For
 * more information on DataONE, see our web site at http://dataone.org.
 *
 *   Copyright 2014
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dataone.tidy.dao;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.dataone.cn.dao.DataSourceFactory;
import org.dataone.cn.dao.SystemMetadataDaoMetacatImpl;
import org.dataone.cn.dao.exceptions.DataAccessException;
import org.dataone.configuration.Settings;
import org.dataone.service.types.v1.Identifier;
import org.dataone.service.types.v1.NodeReference;
import org.dataone.service.types.v1.SystemMetadata;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.PreparedStatementCreator;
import org.springframework.jdbc.core.RowMapper;

/**
 * A class used to access cached Metacat system metadata from associated tables. It relies on the
 * the tables to be created and populated from an outside source before reading and writing to them.
 * 
 * @author cjones
 */
public class SystemMetadataTidyDaoImpl implements SystemMetadataTidyDao {

    private static final Log log = LogFactory.getLog(SystemMetadataTidyDaoImpl.class);

    private JdbcTemplate jdbcTemplate;
    private SystemMetadataDaoMetacatImpl systemMetadataDao;

    // UCSB cached data table configurations
    private static final String UCSB_IDENTIFIER_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.ucsb.identifier");
    private static final String UCSB_SYSMETA_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.ucsb.systemmetadata");
    private static final String UCSB_SM_POLICY_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.ucsb.smreplpolicy");
    private static final String UCSB_SM_STATUS_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.ucsb.smreplstatus");
    private static final String UCSB_ACCESS_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.ucsb.access");

    // UCSB cached data table configurations
    private static final String ORC_IDENTIFIER_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.orc.identifier");
    private static final String ORC_SYSMETA_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.orc.systemmetadata");
    private static final String ORC_SM_POLICY_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.orc.smreplpolicy");
    private static final String ORC_SM_STATUS_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.orc.smreplstatus");
    private static final String ORC_ACCESS_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.orc.access");

    // UCSB cached data table configurations
    private static final String UNM_IDENTIFIER_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.unm.identifier");
    private static final String UNM_SYSMETA_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.unm.systemmetadata");
    private static final String UNM_SM_POLICY_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.unm.smreplpolicy");
    private static final String UNM_SM_STATUS_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.unm.smreplstatus");
    private static final String UNM_ACCESS_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.unm.access");

    // Merged data table configurations
    private static final String MERGED_IDENTIFIER_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.merged.identifier");
    private static final String MERGED_SYSMETA_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.merged.systemmetadata");
    private static final String MERGED_SM_POLICY_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.merged.smreplpolicy");
    private static final String MERGED_SM_STATUS_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.merged.smreplstatus");
    private static final String MERGED_ACCESS_TABLE = Settings.getConfiguration().getString(
            "tidy.dao.merged.access");

    protected static Map<String, String> unmTableMap = new HashMap<String, String>();
    protected static Map<String, String> orcTableMap = new HashMap<String, String>();
    protected static Map<String, String> ucsbTableMap = new HashMap<String, String>();

    protected static Map<String, String> mergedTableMap = new HashMap<String, String>();

    static {
        unmTableMap.put(SystemMetadataDaoMetacatImpl.IDENTIFIER_TABLE, UNM_IDENTIFIER_TABLE);
        unmTableMap.put(SystemMetadataDaoMetacatImpl.SYSMETA_TABLE, UNM_SYSMETA_TABLE);
        unmTableMap.put(SystemMetadataDaoMetacatImpl.SM_POLICY_TABLE, UNM_SM_POLICY_TABLE);
        unmTableMap.put(SystemMetadataDaoMetacatImpl.SM_STATUS_TABLE, UNM_SM_STATUS_TABLE);
        unmTableMap.put(SystemMetadataDaoMetacatImpl.ACCESS_TABLE, UNM_ACCESS_TABLE);

        orcTableMap.put(SystemMetadataDaoMetacatImpl.IDENTIFIER_TABLE, ORC_IDENTIFIER_TABLE);
        orcTableMap.put(SystemMetadataDaoMetacatImpl.SYSMETA_TABLE, ORC_SYSMETA_TABLE);
        orcTableMap.put(SystemMetadataDaoMetacatImpl.SM_POLICY_TABLE, ORC_SM_POLICY_TABLE);
        orcTableMap.put(SystemMetadataDaoMetacatImpl.SM_STATUS_TABLE, ORC_SM_STATUS_TABLE);
        orcTableMap.put(SystemMetadataDaoMetacatImpl.ACCESS_TABLE, ORC_ACCESS_TABLE);

        ucsbTableMap.put(SystemMetadataDaoMetacatImpl.IDENTIFIER_TABLE, UCSB_IDENTIFIER_TABLE);
        ucsbTableMap.put(SystemMetadataDaoMetacatImpl.SYSMETA_TABLE, UCSB_SYSMETA_TABLE);
        ucsbTableMap.put(SystemMetadataDaoMetacatImpl.SM_POLICY_TABLE, UCSB_SM_POLICY_TABLE);
        ucsbTableMap.put(SystemMetadataDaoMetacatImpl.SM_STATUS_TABLE, UCSB_SM_STATUS_TABLE);
        ucsbTableMap.put(SystemMetadataDaoMetacatImpl.ACCESS_TABLE, UCSB_ACCESS_TABLE);

        mergedTableMap.put(SystemMetadataDaoMetacatImpl.IDENTIFIER_TABLE, MERGED_IDENTIFIER_TABLE);
        mergedTableMap.put(SystemMetadataDaoMetacatImpl.SYSMETA_TABLE, MERGED_SYSMETA_TABLE);
        mergedTableMap.put(SystemMetadataDaoMetacatImpl.SM_POLICY_TABLE, MERGED_SM_POLICY_TABLE);
        mergedTableMap.put(SystemMetadataDaoMetacatImpl.SM_STATUS_TABLE, MERGED_SM_STATUS_TABLE);
        mergedTableMap.put(SystemMetadataDaoMetacatImpl.ACCESS_TABLE, MERGED_ACCESS_TABLE);
    }

    public SystemMetadataTidyDaoImpl() {
        this.jdbcTemplate = new JdbcTemplate(DataSourceFactory.getPostgresDataSource());
        this.systemMetadataDao = new SystemMetadataDaoMetacatImpl(
                DataSourceFactory.getPostgresDataSource());
    }

    public SystemMetadataTidyDaoImpl(DataSource dataSource) {
        this.jdbcTemplate = new JdbcTemplate(dataSource);
        this.systemMetadataDao = new SystemMetadataDaoMetacatImpl(dataSource);
    }

    /* (non-Javadoc)
     * @see org.dataone.tidy.dao.SystemMetadataDao#listIdentifiers()
     */
    @Override
    public List<Identifier> listIdentifiers() throws DataAccessException {

        List<Identifier> identifiers = new ArrayList<Identifier>();

        identifiers = jdbcTemplate.query(new PreparedStatementCreator() {

            @Override
            public PreparedStatement createPreparedStatement(Connection conn) throws SQLException {

                // Get a list of unique identifiers via a union of all 3 CN holdings
                StringBuffer sqlStatement = new StringBuffer();
                sqlStatement.append("SELECT guid FROM ");
                sqlStatement.append(UCSB_IDENTIFIER_TABLE);
                sqlStatement.append(" UNION ");
                sqlStatement.append("SELECT guid FROM ");
                sqlStatement.append(ORC_IDENTIFIER_TABLE);
                sqlStatement.append(" UNION ");
                sqlStatement.append("SELECT guid FROM ");
                sqlStatement.append(UNM_IDENTIFIER_TABLE);
                sqlStatement.append(" UNION ");
                sqlStatement.append("SELECT guid FROM ");
                sqlStatement.append(UCSB_SYSMETA_TABLE);
                sqlStatement.append(" UNION ");
                sqlStatement.append("SELECT guid FROM ");
                sqlStatement.append(ORC_SYSMETA_TABLE);
                sqlStatement.append(" UNION ");
                sqlStatement.append("SELECT guid FROM ");
                sqlStatement.append(UNM_SYSMETA_TABLE);

                sqlStatement.append(";");

                log.debug("listIdentifiers() SQL statement: " + sqlStatement.toString());

                PreparedStatement statement = conn.prepareStatement(sqlStatement.toString());
                return statement;
            }

        }, new IdentifierRowMapper());
        return identifiers;
    }

    /* (non-Javadoc)
     * @see org.dataone.tidy.dao.SystemMetadataDao#getSystemMetadata(org.dataone.service.types.v1.Identifier)
     */
    @Override
    public Map<NodeReference, SystemMetadata> getSystemMetadata(Identifier identifier)
            throws DataAccessException {
    	
    	log.info("********** entering TidySysmetaDAOImpl getSysMeta ********"); 
        Map<NodeReference, SystemMetadata> sysMetaMap = new HashMap<NodeReference, SystemMetadata>();

        // Get system metadata for ucsb
        NodeReference ucsbNodeId = new NodeReference();
        ucsbNodeId.setValue(Settings.getConfiguration().getString("tidy.dao.ucsb.nodeid"));
        SystemMetadata ucsbSysMeta = systemMetadataDao.getSystemMetadata(identifier, ucsbTableMap);

        // Get system metadata for orc

        NodeReference orcNodeId = new NodeReference();
        orcNodeId.setValue(Settings.getConfiguration().getString("tidy.dao.orc.nodeid"));
        SystemMetadata orcSysMeta = systemMetadataDao.getSystemMetadata(identifier, orcTableMap);

        // Get system metadata for unm
        NodeReference unmNodeId = new NodeReference();
        unmNodeId.setValue(Settings.getConfiguration().getString("tidy.dao.unm.nodeid"));
        SystemMetadata unmSysMeta = systemMetadataDao.getSystemMetadata(identifier, unmTableMap);

        // populate the map
        if (ucsbSysMeta != null) {
            sysMetaMap.put(ucsbNodeId, ucsbSysMeta);
        }
        if (orcSysMeta != null) {
            sysMetaMap.put(orcNodeId, orcSysMeta);
        }
        if (unmSysMeta != null) {
            sysMetaMap.put(unmNodeId, unmSysMeta);
        }
        log.info("TidySysmetaDAOImpl available sysmetas for pid: " + identifier.getValue() +
				" : " + sysMetaMap.size());
        
        if (sysMetaMap.isEmpty()) {
        	log.error("TidySysmetaDAOimpl: No SystemMetadata available for pid: " + identifier.getValue());
        	StringBuffer sb = new StringBuffer();
        	if (ucsbTableMap != null) {
        		for (Entry<String, String> n : ucsbTableMap.entrySet()) {
        			sb.append(String.format("%s=%s: ",n.getKey(), n.getValue()));
        		}
        		log.warn(    "(sample table map: " + sb.toString());
        	} else {
        		log.error(   "ucsb tablemap is null!!");
        	}
        } else {
        	if (log.isDebugEnabled()) {
        		for (Entry<NodeReference,SystemMetadata> n : sysMetaMap.entrySet()) {
        			log.debug(String.format("%s = %s", n.getKey().getValue(), n.getValue()));
        		}
        	}
        }
        return sysMetaMap;
    }

    /**
     * @see org.dataone.tidy.dao.SystemMetadataDao#saveSystemMetadata(org.dataone.service.types.v1.SystemMetadata)
     */
    @Override
    public void saveSystemMetadata(SystemMetadata systemMetadata) throws DataAccessException {

        // save the merged system metadata to the merged system metadata tables
        systemMetadataDao.saveSystemMetadata(systemMetadata, mergedTableMap);

    }

    /**
     * An internal class that maps guid rows to Identifier objects via the RowMapper interface
     */
    public static final class IdentifierRowMapper implements RowMapper<Identifier> {

        @Override
        public Identifier mapRow(ResultSet resultSet, int rowNumber) throws SQLException {
            Identifier identifier = new Identifier();
            identifier.setValue(resultSet.getString("guid"));

            return identifier;
        }

    }
}
